#!/usr/bin/env python3
import pickle, scipy, sys,  fileinput, gc, os
from networkx import *
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import mstats, rankdata
import seaborn as sns
import pandas as pd
import time
from copy import deepcopy
from scipy.stats.mstats import f_oneway as aov
from subprocess import Popen
from statsmodels.stats.multicomp import pairwise_tukeyhsd as TukeyHSD
from statsmodels.stats.libqsturng import psturng

try:
    import community
except:
    community_installed = False
    print('https://github.com/taynaud/python-louvain is a requirement for finding communites')
else:
    import community
    community_installed = True
gc.enable()
######################################################################################
###############
## basic function library
def read_file(tempFile,linesOraw,quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)


def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))


def process_dir(in_dir):
    ## process the output dir
    in_dir = os.path.join(in_dir)
    if not os.path.isdir(in_dir):
        os.makedirs(in_dir)
    return(in_dir)
############################################################################################
## this function was adopted from emre's stackoverflow answer found here:
## https://stackoverflow.com/questions/7450957/how-to-implement-rs-p-adjust-in-python
def correct_pvalues_for_multiple_testing(pvalues, correction_type = "Benjamini-Hochberg"):                
    """                                                                                                   
    consistent with R - print correct_pvalues_for_multiple_testing([0.0, 0.01, 0.029, 0.03, 0.031, 0.05, 0.069, 0.07, 0.071, 0.09, 0.1]) 
    """
    from numpy import array, empty
    pvalues = array(pvalues)
    n = int(pvalues.shape[0])
    new_pvalues = empty(n)
    if correction_type == "Bonferroni":
        new_pvalues = n * pvalues
    elif correction_type == "Bonferroni-Holm":
        values = [ (pvalue, i) for i, pvalue in enumerate(pvalues) ]
        values.sort()
        for rank, vals in enumerate(values):
            pvalue, i = vals
            new_pvalues[i] = (n-rank) * pvalue
    elif correction_type == "Benjamini-Hochberg":
        values = [ (pvalue, i) for i, pvalue in enumerate(pvalues) ]
        values.sort()
        values.reverse()
        new_values = []
        for i, vals in enumerate(values):
            rank = n - i
            pvalue, index = vals
            new_values.append((n/rank) * pvalue)
        for i in range(0, int(n)-1):
            if new_values[i] < new_values[i+1]:
                new_values[i+1] = new_values[i]
        for i, vals in enumerate(values):
            pvalue, index = vals
            new_pvalues[index] = new_values[i]
    return new_pvalues

def get_tukey(vals, groups):
    try:
        res = TukeyHSD(vals, groups)
    except:
        return([['group_1','group_2',"TukeyHSD_p_values"],["NA","NA","1"]])
    print('\n\n',res)
    first_group_vector = ['group_1']
    second_group_vector = ['group_2']
    for i in range(res.groupsunique.shape[0]):
        for j in range(i,res.groupsunique.shape[0]):
            if i != j:
                first_group_vector.append(res.groupsunique[i])
                second_group_vector.append(res.groupsunique[j])

    p_vals = psturng(np.abs(res.meandiffs / res.std_pairs), len(res.groupsunique), res.df_total)
    p_values = ['TukeyHSD_p_values'] + np.array(p_vals,dtype=str).tolist()
    for p in range(0,len(p_values)):
        if p_values[p]=='0.001':
            p_values[p] = '<0.001'

    print(p_values)
    out_table = np.transpose(np.vstack((first_group_vector, second_group_vector, p_values)))
    return (out_table)
############################################################################################
##############################################################
import argparse
parser = argparse.ArgumentParser()


parser.add_argument("-community_table",
                    help="the table annotating the ")
parser.add_argument("-id_subset",
                    help="a file that contains all of the gene IDs in the current module to be plotted")
parser.add_argument("-z_score_table",
                    help="the table of cluster level Z-scores")
parser.add_argument("-module_size_cutoff",
                    help="how big is the smallest module that we should analyze? (default = 15)",
                    default = 15)
parser.add_argument("-species",
                    help="for pathway analysis, we'll need the species (default = hsapiens)",
                    default = "hsapiens")
parser.add_argument("-name",
                    help="if you only have a sinlge input list, you can name it here",
                    default = "community")
parser.add_argument('-annotation_dict','-ad',
                    dest='symbol_def_dict',
                    type=str)
parser.add_argument("-out_dir",
                    help="the output directory")


##############################################################

def do_gprofiler(ids, out_dir, gprofiler_out_dir, background, species, name = 'community'):
    bg_file = os.path.join(out_dir,'background.txt')
    make_file('\n'.join(background), bg_file)
    com_file = os.path.join(out_dir,'community_ids.txt')
    make_file('\n'.join(ids), com_file)
    gprofiler_out_file = os.path.join(gprofiler_out_dir,name+'_gprofiler.txt')
    cmd('pyminer_gprofile.py -i '+com_file+' -b '+bg_file+' -species '+species+' -o '+gprofiler_out_file)
    return


def get_gene_annotations(ids, symbol_dict, def_dict):
    out_anno = []
    for i in range(len(ids)):
        temp_node = ids[i]
        if len(symbol_dict[temp_node]) == 0:
            out_anno.append([temp_node, "None", "None"])
        else:
            for j in range(len(symbol_dict[temp_node])):
                out_anno.append([temp_node, symbol_dict[temp_node][j], def_dict[temp_node][j]])
                #print(temp_node, symbol_dict[temp_node][j], def_dict[temp_node][j])
    return(out_anno)


def pr_annotation(in_table, pr_dict):
    if pr_dict == None:
        return(in_table)
    pr_vect = []
    for i in range(0,len(in_table)):
        ## first col is gene
        in_table[i].append(pr_dict[in_table[i][0]])
        pr_vect.append(pr_dict[in_table[i][0]])
    ## sort the table based on highest page rank
    pr_vect_order = np.argsort(np.array(pr_vect)).tolist()[::-1]
    in_table = np.array(in_table)
    # print(in_table)
    # print(np.shape(in_table))
    # print(pr_vect_order)
    in_table = in_table[pr_vect_order,:].tolist()
    return(in_table)


def get_med_table(z_array_header, z_array_subset, name, out_dir):
    #print(z_array_subset)
    temp_gmean = np.median(z_array_subset, axis = 0).tolist()
    out_table = [['id',name]]
    for i in range(len(z_array_header)):
        out_table.append([z_array_header[i], temp_gmean[i]])
    write_table(out_table, os.path.join(out_dir, 'median_module_usage.tsv'))
    return()


def analyze_subset(ids, 
                   z_array,
                   z_array_header,
                   id_list,
                   id_hash, 
                   out_dir,
                   gprofiler_out_dir,
                   species,
                   symbol_dict,
                   def_dict,
                   name = 'community',
                   background = None,
                   pr_dict = None):
    ## if there isn't a background, just use everything in the dataset
    if background == None:
        background = id_list

    ## first thing to do is catelogue the index of the ids from the z-array
    temp_idxs = sorted([id_hash[temp_id] for temp_id in ids])
    
    z_subset_array = z_array[temp_idxs,:]
    ## z_subset_array_rank = rankdata(z_subset_array, method = 'min')

    ## get the median for each group
    get_med_table(deepcopy(z_array_header), z_subset_array, name, out_dir)

    ## strip the header of 'sample_group_'
    for i in range(0,len(z_array_header)):
        z_array_header[i]=z_array_header[i].replace('sample_group_','')
    z_subset_lists = [z_subset_array[:,i].tolist() for i in range(z_subset_array.shape[1])]
    temp_aov_f, temp_aov_p = aov(z_subset_lists)
    print('\t\tnominal F/p',temp_aov_f, temp_aov_p)
    ## now plot the results

    #####################
    ## we will use the actual data for display & rank transformed Z-scores for Tukey HSD

    # z_subset_rank = pd.DataFrame(z_subset_array_rank)
    # z_subset_rank.columns = z_array_header
    # z_subset_rank = z_subset_rank.melt(var_name='groups', value_name='z-scores')


    z_subset = pd.DataFrame(z_array[temp_idxs,:])
    z_subset.columns = z_array_header
    z_subset = z_subset.melt(var_name='groups', value_name='z-scores')

    plt.clf()
    sns.violinplot(x="groups",y="z-scores", data = z_subset)
    plt.savefig(os.path.join(out_dir,"group_z_scores.png"),
                dpi=300,
                bbox_inches='tight')

    ## do the post-hoc tests
    tukey_table = get_tukey(z_subset["z-scores"],z_subset["groups"])
    write_table(tukey_table, os.path.join(out_dir,'TukeyHSD.tsv'))

    ## write the annotation file in the output directory
    temp_gene_annotations = get_gene_annotations(ids, symbol_dict, def_dict)
    temp_gene_annotations = pr_annotation(temp_gene_annotations, pr_dict = pr_dict)
    write_table(temp_gene_annotations, os.path.join(out_dir,'community_ids_annotated.tsv'))


    ## do the gprofiler analysis of this module
    do_gprofiler(ids, out_dir, gprofiler_out_dir, background, species, name = name)

    cmd("plot_network.py -node_color community -subset_nodes "+os.path.join(out_dir,"community_ids.txt")+" -out_dir "+out_dir+" -graph "+os.path.join(os.path.dirname(os.path.dirname(out_dir)),"large_comps.graphpkl")+" -no_default -dont_save")
    

    return (temp_aov_f, temp_aov_p, z_subset, temp_gene_annotations)


def combine_community_level_global_stats(out_dir, 
                                         com_names,
                                         f,
                                         p,
                                         p_correct):
    output = [['community_names', "F_statistic","nominal_p","BH_corrected_p"]]
    for i in range(len(com_names)):
        output.append([com_names[i],
                       f[i],
                       p[i],
                       p_correct[i]])
    write_table(output,os.path.join(out_dir,'global_statistics.tsv'))
    return



def make_pr_dict(community_table):
    pr_dict = {}
    ## this table has a header, so start at 1
    for i in range(1,community_table.shape[0]):
        pr_dict[community_table[i,0]] = float(community_table[i,2])
    return(pr_dict)


def do_community_level_analyses(z_array, 
                                z_array_header,
                                community_file, 
                                id_list, ## this is for the z-score table
                                id_hash, ## for z-score table
                                out_dir,
                                species,
                                symbol_dict,
                                def_dict,
                                module_size_cutoff = 15):

    ## first go through getting all of the modules & figure out how big they are
    community_table = np.array(read_table(community_file))
    pr_dict = make_pr_dict(community_table)

    ## catelogue all of the communities & their members
    all_communities = set(community_table[1:,1].tolist())
    community_dict = {com:[] for com in all_communities}
    for i in range(1,community_table.shape[0]):
        temp_com = community_table[i,1]
        temp_com_list = community_dict[temp_com]
        temp_com_list.append(community_table[i,0])
        community_dict[temp_com] = temp_com_list

    ## get the final list of all the communities to include in the actual analysis
    final_analysis_coms = []
    for com in all_communities:
        if len(community_dict[com]) >= module_size_cutoff:
            final_analysis_coms.append(com)

    ## catelogue all of the genes included for the appropriate background in doing pathway analysis
    all_genes_for_gprofiler_bg = []
    for com in final_analysis_coms:
        all_genes_for_gprofiler_bg += community_dict[com]

    ## now go through each module & run the analyses
    out_dir = process_dir(out_dir)
    gprofiler_out_dir = process_dir(os.path.join(out_dir,'gprofiler'))
    stats_f = []
    stats_p = []
    z_subset_lists = []
    community_gene_annotations = []
    com_out_dir_dict = {}
    for com in final_analysis_coms:
        temp_out_dir = process_dir(os.path.join(out_dir,com))
        com_out_dir_dict[com] = temp_out_dir
        print('\tworking on',com,len(community_dict[com]))
        temp_stats_results = analyze_subset(community_dict[com], 
                                            z_array,
                                            z_array_header,
                                            id_list,
                                            id_hash, 
                                            temp_out_dir,
                                            gprofiler_out_dir,
                                            species,
                                            symbol_dict,
                                            def_dict,
                                            background = all_genes_for_gprofiler_bg,
                                            name = com,
                                            pr_dict = pr_dict)
        stats_f.append(temp_stats_results[0])
        stats_p.append(temp_stats_results[1])
        z_subset_lists.append(temp_stats_results[2])
        community_gene_annotations.append(temp_stats_results[3])
    ## correct for multipel comparisons
    stats_p_correct = correct_pvalues_for_multiple_testing(stats_p)
    combine_community_level_global_stats(out_dir, 
                                         final_analysis_coms,
                                         stats_f,
                                         stats_p,
                                         stats_p_correct)

    cmd('combine_gprofiler_results.py -i '+gprofiler_out_dir+' -o '+out_dir)


    return


def do_module_analysis(z_score_table,
                       out_dir,
                       species = 'hsapiens',
                       symbol_def_dict = None,
                       community_file = None,
                       id_subset = None,
                       module_size_cutoff = 15,
                       name = 'community'):
    """
    Takes in either the community table & makes subdirs in the outdir for each community larger than the cutoff size
     OR 
    a subset of ids & plots the result directly in the out dir.

    """
    if (community_file == None) and (id_subset == None):
        sys.exit("need either the community table or a subset of ids to work on")
    if (community_file != None) and (id_subset != None):
        sys.exit("can only work with either a community table or an id subset")
    
    ## no matter what, we'll need to use the z-score table, so we can process it now
    z_table = np.array(read_table(z_score_table))
    z_array_header = z_table[0,1:].tolist()
    id_list = z_table[1:,0].tolist()
    id_hash = {key:value for value, key in enumerate(id_list)}
    z_array = np.array(z_table[1:,1:],dtype = np.float32)

    ## import the annotation dict
    symbol_dict, def_dict = import_dict(symbol_def_dict)

    ## if we are doing the full community analysis
    if (community_file != None) and (id_subset == None):
        print("doing community analysis")
        do_community_level_analyses(z_array, 
                                    z_array_header,
                                    community_file, 
                                    id_list, ## this is for the z-score table
                                    id_hash, ## for z-score table
                                    out_dir,
                                    species,
                                    symbol_dict,
                                    def_dict,
                                    module_size_cutoff = module_size_cutoff)

    ## if there is just a subset of genes that we want to work with
    if (community_file == None) and (id_subset != None):
        print("doing enrichment analysis on your defined gene-set")
        ## read it in
        subset_ids = read_file(id_subset,'lines')
        analyze_subset(subset_ids, 
                        z_array,
                        z_array_header,
                        id_list,
                        id_hash, 
                        out_dir,
                        out_dir,
                        species,
                        symbol_dict, def_dict,
                        name = name)
    return



if __name__=="__main__":
    args = args = parser.parse_args()
    do_module_analysis(args.z_score_table, 
                       args.out_dir,
                       species = args.species,
                       community_file = args.community_table,
                       id_subset = args.id_subset,
                       symbol_def_dict = args.symbol_def_dict,
                       module_size_cutoff = args.module_size_cutoff,
                       name = args.name)



#!/usr/bin/env python3

##import dependency libraries
import sys,time,glob,os,pickle,fileinput
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
import argparse
#import pandas as pd
##############################################################
## basic function library
def read_file(tempFile,linesOraw='lines',quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def get_file_path(in_path):
    in_path = in_path.split('/')
    in_path = in_path[:-1]
    in_path = '/'.join(in_path)
    return(in_path+'/')


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)
    

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)



##############################################################



parser = argparse.ArgumentParser()

## global arguments
parser.add_argument(
	'-means','-m','-mean',
	dest='means',
	type=str)

parser.add_argument(
	'-significance','-sig','-anovas','-anova', '-aov',
	dest='sig',
	type=str)

parser.add_argument(
	'-out','-o',
    help = "output directory",
	dest='out',
	type=str)

parser.add_argument(
	'-annotation_dict','-ad',
	dest='symbol_def_dict',
	type=str)


args = parser.parse_args()
##################################################################

## import the annotation dict
symbol_dict, def_dict = import_dict(args.symbol_def_dict)

## hyperparameters
sig_cutoff = 0.05
q_cutoff = 0.5
## the percent of the largest distance to second to consider for marker genes
## this gives you highly expressed differentially expressed genes
percentile = 0.9


k_group_means_file = np.array(read_table(args.means))
id_list = k_group_means_file[1:,0].tolist()
group_names = k_group_means_file[0,1:].tolist()

if len(group_names)<3:
	sys.exit('need at least 3 groups')

###########################
## get the means array
k_group_means = np.array(k_group_means_file[1:,1:],dtype=float)
k_group_min = np.min(k_group_means,axis=1)
k_group_max = np.max(k_group_means,axis=1)
k_group_range = k_group_max - k_group_min
## get the distance from max to second max
dist_max_to_second = np.zeros((len(id_list)))
q_value_vect = np.zeros((len(id_list)))
for i in range(0,len(id_list)):
	temp_sorted = np.sort(k_group_means[i,:])[::-1]
	dist_max_to_second[i]=temp_sorted[0]-temp_sorted[1]
	q_value_vect[i] = dist_max_to_second[i]/k_group_range[i]

sorted_dist_to_max = np.sort(dist_max_to_second[:],kind="mergesort")
index_for_cutoff = int(len(id_list)*percentile)
dist_max_to_second_cutoff = sorted_dist_to_max[index_for_cutoff]

##########################
## 
BH_corrected_aov_file = np.array(read_table(args.sig))
BH_corrected_aov = np.array(BH_corrected_aov_file[1:,3],dtype=float)

##
## only consider significant genes
sig_bool = BH_corrected_aov<sig_cutoff
q_bool = q_value_vect>q_cutoff
dist_to_second_max_bool = dist_max_to_second>dist_max_to_second_cutoff
print('\n\n\nsig_bool',sig_bool,sum(sig_bool))
print('q_bool',q_bool,sum(q_bool))
print('dist_to_second_max_bool',dist_to_second_max_bool,sum(dist_to_second_max_bool))
sig_q_and_dist = (np.array(sig_bool,dtype=int) + np.array(q_bool,dtype=int) + np.array(dist_to_second_max_bool,dtype=int))==3
print('\ncombined\n',sig_q_and_dist)


print('found',sum(sig_q_and_dist),'highly expressed, somewhat "exclusive" marker genes')

print('making the output table...')

out_file = [['gene','gene_symbol','gene_def', 'aov_BH_p', 'range', 'dist_from_max_to_second_max', 'q_value', 'group']]


def get_group(i):
	global sig_q_and_dist, group_names, k_group_max, k_group_means
	if not sig_q_and_dist[i]:
		return("None")
	else:
		index = int(np.where(k_group_means[i] == k_group_max[i])[0])
		return(group_names[index])

def process_symbol(in_symbol_list,ensg):
	if len(in_symbol_list)==0:
		return(ensg)
	else:
		return('_'.join(in_symbol_list))

def process_def(in_def):
	if len(in_def)==0:
		return("NA")
	else:
		return(in_def[0])

for i in range(0,len(id_list)):
    ## in the case of entrez IDs
    try:
        int(float(id_list[i]))
    except:
        temp_id = id_list[i]
    else:
        temp_id = str(int(float(id_list[i])))
    temp_line = []
    temp_line.append(temp_id)
    temp_line.append(process_symbol(symbol_dict[temp_id],temp_id))
    temp_line.append(process_def(def_dict[temp_id]))
    temp_line.append(BH_corrected_aov[i])
    temp_line.append(k_group_range[i])
    temp_line.append(dist_max_to_second[i])
    temp_line.append(q_value_vect[i])
    temp_line.append(get_group(i))
    out_file.append(temp_line)

## get the input for plot_subset...
plot_subset_input = []
for i in range(0,len(out_file)):
	if out_file[i][-1]!="None":
		plot_subset_input.append([out_file[i][0],out_file[i][1]])

subset_input_file = args.out+"/subset_input.txt"
write_table(plot_subset_input,subset_input_file)
write_table(out_file,args.out+'/marker_gene_annotations.tsv')

#cmd("plot_subset.py -i ""-plot_subset "+subset_input_file)






#!/usr/bin/env python3
import time
import os
import pickle
import sys
import json
import time
import argparse
from subprocess import Popen
from copy import deepcopy

# Python 2/3 adaptability
try:
    from urllib.parse import urlparse, urlencode
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError
except ImportError:
    from urlparse import urlparse
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError
######################################################################
##############################################################
## basic function library
def read_file(tempFile,linesOraw='lines',quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim, num_type = float):
    #print(num_type)
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                if num_type == float:
                    lines[i][j]=float(lines[i][j])
                elif num_type == int:
                    lines[i][j]=int(float(lines[i][j]))
                else:
                    lines[i][j]=num_type(lines[i][j])
    return(lines)


def get_file_path(in_path):
    in_path = in_path.split('/')
    in_path = in_path[:-1]
    in_path = '/'.join(in_path)
    return(in_path+'/')


def read_table(file, sep='\t',num_type=float):
    return(make_table(read_file(file,'lines'),sep,num_type=num_type))
    

def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)
    

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)


def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()


def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)

##############################################################


######################################################################
## 

######################################################################

## the below block of code was built from:
## https://github.com/Ensembl/ensembl-rest/wiki/Example-Python-Client
class EnsemblRestClient(object):
    def __init__(self, server='http://rest.ensembl.org', reqs_per_sec=15):
        self.server = server
        self.reqs_per_sec = reqs_per_sec
        self.req_count = 0
        self.last_req = 0
    def perform_rest_action(self, endpoint, hdrs=None, params=None):
        if hdrs is None:
            hdrs = {}
        if 'Content-Type' not in hdrs:
            hdrs['Content-Type'] = 'application/json'
        if params:
            endpoint += '?' + urlencode(params)
        data = None
        print(endpoint)
        # check if we need to rate limit ourselves
        if self.req_count >= self.reqs_per_sec:
            delta = time.time() - self.last_req
            if delta < 1:
                time.sleep(1 - delta)
            self.last_req = time.time()
            self.req_count = 0
        try:
            request = Request(self.server + endpoint, headers=hdrs)
            response = urlopen(request)
            content = response.read()
            if content:
                data = json.loads(content)
            self.req_count += 1
        except HTTPError as e:
            # check if we are being rate limited by the server
            if e.code == 429:
                if 'Retry-After' in e.headers:
                    retry = e.headers['Retry-After']
                    time.sleep(float(retry))
                    self.perform_rest_action(endpoint, hdrs, params)
            else:
                sys.stderr.write('Request failed for {0}: Status code: {1.code} Reason: {1.reason}\n'.format(endpoint, e))
        return data
    def get_variants(self, species, symbol):
        genes = self.perform_rest_action(
            endpoint='/xrefs/symbol/{0}/{1}'.format(species, symbol), 
            params={'object_type': 'gene'}
        )
        if genes:
            stable_id = genes[0]['id']
            variants = self.perform_rest_action(
                '/overlap/id/{0}'.format(stable_id),
                params={'feature': 'variation'}
            )
            return variants
        return None


    def get_regulatory_elements_from_region(self, contig, start, end, feat = ["regulatory","other_regulatory", "motif"]):
        out_dict = {}
        for f in feat:
            endpoint = "/overlap/region/human/"+contig+":"+start+".."+end
            endpoint += "?feature="+f
            print('\n\n\n',endpoint)
            temp_locus_response = self.perform_rest_action(endpoint)#, params = "feature=gene")
            print(temp_locus_response,'\n\n\n')
            out_dict[f] = temp_locus_response
        return(out_dict)

######################################################################


def run(species, symbol):
    ## help with resful: http://rest.ensembl.org/?content-type=text/html
    ## http://rest.ensembl.org/variation/human/rs1199776527
    ## http://rest.ensembl.org/overlap/region/human/7:140924917..140924917?feature=gene
    ## http://rest.ensembl.org/info/variation/consequence_types/
    ###### eqtl with result
    ## http://rest.ensembl.org/eqtl/variant_name/human/rs7683255
    ###### eqtl without result
    ## http://rest.ensembl.org/eqtl/variant_name/human/rs1310656119 
    client = EnsemblRestClient()
    variants = client.get_variants(species, symbol)
    if variants:
        for v in variants:
            print('{seq_region_name}:{start}-{end}:{strand} ==> {id} ({consequence_type})'.format(**v))

######################################################################

class variant_obj():
    def __init__(self,client,rsID):
        self.id = rsID
        self.var_details = {}
        self.genes = []## these are in Ensembl format
        self.nearest_genes = []## this is a list in case there are several genes that overlap with the variant locus
        self.all_locus_response = []
        self.mapped_loci = []
        self.prot_muts = []
        self.trans_muts = []
        ## general mutation info
        self.var_is_bad = False
        self.mutation_type = None
        ## regulatory element info
        self.regulatory_elements = []
        self.regulatory = []
        self.other_regulatory = []
        self.motif = []
        ## tad info
        self.tad_boundaries_crossed = 0
        self.tads = []
        self.tad_annotations = []
        ## gene info
        self.detailed_gene_info = []
        self.gene_symbols = []## mapped through gProfiler to symbols from Ensembl
        self.symbol_to_gene_lookup = {}## this is for the reverse lookup after finding gene symbol information to go back to the gene that that symbol come from
        ## search term info
        self.gene_search_term_comentions = []## search of the gene symbols that mapped from this variant to the search terms
        self.best_gene_search_term_match = []
        ## eqtls
        self.eqtls={}
        ## scRNAseq mapping
        self.scRNAseq_ids = []
        self.get_direct_cause(client,self.id)


    def check_bad(self, in_consequence):
        types_of_mutations_acceptable = ['3_prime_UTR_variant','5_prime_UTR_variant','coding_sequence_variant','feature_truncation','frameshift_variant','incomplete_terminal_codon_variant','inframe_deletion','inframe_insertion','mature_miRNA_variant','missense_variant','NMD_transcript_variant','non_coding_transcript_exon_variant','protein_altering_variant','splice_acceptor_variant','splice_donor_variant','splice_region_variant','start_lost','start_retained_variant','stop_gained','stop_lost','stop_retained_variant','transcript_ablation','transcript_amplification']
        if in_consequence in types_of_mutations_acceptable:
            return(True)
        else:
            return(False)


    def is_bad_mutation(self,var_details):
        if "most_severe_consequence" in var_details:
            self.mutation_type = var_details["most_severe_consequence"]
            if self.check_bad(self.mutation_type):
                return(True)
        return(False)


    def set_all_loci(self, var_details):
        genes = []
        if var_details == {}:
            ## take care of the null case in which it wasn't mappable
            self.mapped_loci.append({"contig":"NA", "start":0, "end":0})
        else:
            ## one approach is to get the intervals and look that up 
            for locus in var_details["mappings"]:
                #print(locus)
                if locus["assembly_name"]=="GRCh38":
                    temp_contig = locus["seq_region_name"]
                    temp_start = locus["start"]
                    if "end" in locus:
                        temp_end = locus["end"]
                    else:
                        temp_end = deepcopy(temp_start)
                    interval = [temp_start,temp_end]
                    start = int(min(interval))
                    end = int(max(interval))
                    self.mapped_loci.append({"contig":temp_contig, "start":start, "end":end})
                    if temp_contig[:3] != "chr":
                        self.mapped_loci.append({"contig":"chr"+temp_contig, "start":start, "end":end})
                    else:
                        self.mapped_loci.append({"contig":temp_contig[3:], "start":start, "end":end})
        print("\n\n",self.mapped_loci,"\n\n")
        return


    def ensp_to_ensg(self, client, ensp_list=[]):
        if len(ensp_list)==0:
            return([])
        enst_list = []
        for ensp in ensp_list:
            #https://rest.ensembl.org/documentation/info/overlap_translation
            temp_response = client.perform_rest_action("/overlap/translation/"+ensp+"?feature=transcript_variation;type="+self.mutation_type+";content-type=application/json")
            if temp_response is None:
                pass
            else:
                for element in temp_response:
                    if element['id'] == self.id:
                        enst_list.append(element["Parent"])
                        for key, value in element.items():
                            print(key, '\n\t',value)
        enst_list = sorted(list(set(enst_list)))
        print(enst_list)
        ensg_list = self.enst_to_ensg(client, enst_list)
        print(ensg_list)
        return(ensg_list)


    def enst_to_ensg(self, client, enst_list=[]):
        if len(enst_list)==0:
            return([])
        ensg_list = []
        for enst in enst_list:
            temp_response = client.perform_rest_action("/lookup/id/"+enst+"?expand=1;content-type=application/json?feature=Parent")
            if temp_response is None:
                pass
            else:
                ensg_list.append(temp_response["Parent"])
        ensg_list = sorted(list(set(ensg_list)))
        print(ensg_list)
        #sys.exit()
        return(ensg_list)


    def get_var_recoder(self, client, var_details):
        print(var_details)
        self.var_details = var_details
        #### this is to replace get_gene_from_var in only the aspect of variant annotation if var_is_bad
        ## We will need to set nearest gene as well as var details, and return the list genes that are annotated as causal (in ENSG format)
        endpoint = "/variant_recoder/human/"+self.id+"?content-type=application/json"
        temp_locus_response = client.perform_rest_action(endpoint)#[0]## zero because this is a list, that only has one element because we're only requesting one variant at a time
        if temp_locus_response is None:
            print("\n\nSOMETHING WENT WRONG WITH VARIANT ANNOTATION\ndefaulting back to original get_gene_from_var function that will simply set it to all genes with overlap")
            return(self.get_gene_from_var(client, var_details))
        if len(temp_locus_response) == 0:
            print("\n\nSOMETHING WENT WRONG WITH VARIANT ANNOTATION\ndefaulting back to original get_gene_from_var function that will simply set it to all genes with overlap")
            return(self.get_gene_from_var(client, var_details))
        ## get all of the proteins and transcripts that this variant changes
        else:
            ## we got a successful reply
            print("we got a successful annotation:",len(temp_locus_response),"results")
            for key, value in temp_locus_response[0].items():
                print(key)
                print("\t",value)
            ensp_list = []
            enst_list = []
            for resp in temp_locus_response:## each snp
                print('\n\n',resp.keys())
                if "hgvsp" in resp:
                    for element in resp["hgvsp"]:
                        self.prot_muts.append(element)
                        if element[:4]=="ENSP":
                            temp_element = element.split(':')
                            temp_element = temp_element[0].split('.')
                            ## after the colon is the actual mutation annotation
                            ensp_list.append(temp_element[0])
                if "hgvsc" in resp:
                    for element in resp["hgvsc"]:
                        self.trans_muts.append(element)
                        if element[:4]=="ENST":
                            ## after the colon is the actual mutation annotation
                            temp_element = element.split(':')
                            temp_element = temp_element[0].split('.')
                            enst_list.append(temp_element[0])
            print(self.prot_muts)
            print(self.trans_muts)
            print(ensp_list)
            print(enst_list)
            if len(ensp_list)>0:
                temp_ensg = self.ensp_to_ensg(client, ensp_list)
                print("protein coding mutations in:")
                print('\t',temp_ensg)
            else:
                temp_ensg = self.enst_to_ensg(client, enst_list)
                print("transcript variant in:")
                print('\t',temp_ensg)
            #sys.exit()
            self.nearest_genes = temp_ensg
            if len(temp_ensg)==0:
                return(self.get_gene_from_var(client, var_details))
            else:
                return(temp_ensg)
            ## "/lookup/id/ENST00000496384?expand=1;content-type=application/json?feature=Parent"


    def get_vep_res(self, client, var_details):
        ## http://rest.ensembl.org/vep/human/id/rs1504215?content-type=application/json
        #print(var_details)
        self.var_details = var_details
        #### this is to replace get_gene_from_var in only the aspect of variant annotation if var_is_bad
        ## We will need to set nearest gene as well as var details, and return the list genes that are annotated as causal (in ENSG format)
        endpoint = "/vep/human/id/"+self.id+"?content-type=application/json"
        response = client.perform_rest_action(endpoint)
        if response is None:
            print("\n\ngot a weird response from VEP. Going to use the old overlap approach\n\n")
            return(self.get_gene_from_var(client, var_details))
        #############################
        ## if we got a valid response....
        all_bad_genes = []
        ## this endpoint can take multiple variants, but we're doing it one at a time for the sake of staying sane
        response = response[0]
        ## now look at the transcript_consequences
        for trans_result in response["transcript_consequences"]:
            for consequence in trans_result["consequence_terms"]:
                if self.check_bad(consequence):
                    print(self.id+":\t",trans_result["gene_id"],":",consequence)
                    all_bad_genes.append(trans_result["gene_id"])
        ensg_list = sorted(list(set(all_bad_genes)))
        self.nearest_genes = ensg_list
        print(self.nearest_genes)
        return(ensg_list)


    def get_gene_from_var(self, client, var_details):
        print(var_details)
        self.var_details = var_details
        genes = []
        self.all_locus_response = []
        ## one approach is to get the intervals and look that up 
        for locus in var_details["mappings"]:
            #print(locus)
            if locus["assembly_name"]=="GRCh38":
                temp_contig = locus["seq_region_name"]
                temp_start = locus["start"]
                if "end" in locus:
                    temp_end = locus["end"]
                else:
                    temp_end = deepcopy(temp_start)
                interval = [temp_start,temp_end]
                start = str(min(interval))
                end = str(max(interval))
                endpoint = "/overlap/region/human/"+temp_contig+":"+start+".."+end+"?feature=gene"
                print(endpoint)
                temp_locus_response = client.perform_rest_action(endpoint)#, params = "feature=gene")
                self.all_locus_response.append(temp_locus_response)
                print('\n',temp_locus_response)
                if temp_locus_response is not None:
                    for entry in temp_locus_response:
                        if "gene_id" in entry:
                            genes.append(entry["gene_id"])
                            self.nearest_genes.append(entry["gene_id"])
        genes = sorted(list(set(genes)))
        return(genes)


    def get_mutation_type(self,client,variant, tries = 3):
        endpoint = "/variation/human/"+variant
        var_details = client.perform_rest_action(endpoint = endpoint)
        if var_details == None and tries > 0:
            time.sleep(.25)
            self.get_mutation_type(client, variant, tries = tries-1)
            return
        if var_details == None:
            var_details = {}
        print(var_details)
        self.var_is_bad = self.is_bad_mutation(var_details)
        print(self.var_is_bad)
        if self.var_is_bad:
            print("getting vep results")
            self.genes = self.get_vep_res(client, var_details)
            #self.genes = self.get_var_recoder(client,var_details)
            #self.genes = self.get_gene_from_var(client,var_details)
        else:
            self.genes = []
        ## set the loci regardless
        self.set_all_loci(var_details)
        return


    # def get_regulatory_elements_for_var(self, contig, start, end, feat = ["regulatory","other_regulatory", "motif"]):## used to include "peak" as well, but it was too cluttered
    #     out_dict = {}
    #     for f in feat:
    #         endpoint = "/overlap/region/human/"+contig+":"+start+".."+end
    #         endpoint += "?feature="+f
    #         print('\n\n\n',endpoint)
    #         temp_locus_response = client.perform_rest_action(endpoint)#, params = "feature=gene")
    #         print(temp_locus_response,'\n\n\n')
    #     return(temp_locus_response)


    def get_regulatory_elements_for_var(self):
        client = EnsemblRestClient()
        self.regulatory_elements = []
        for locus in self.mapped_loci:## dictionary object with contig, start, end
            ### because Ensembl now accepts "chr" leader, we'll not include it for the sake of non-duplication
            if locus["contig"][:3] != "chr":
                temp_response = client.get_regulatory_elements_from_region(locus["contig"], str(locus["start"]), str(locus["end"]))
                print(temp_response)
                if temp_response is not None:
                    if temp_response["regulatory"] is not None:
                        self.regulatory += temp_response["regulatory"]
                    if temp_response["other_regulatory"] is not None:
                        self.other_regulatory += temp_response["other_regulatory"]
                    if temp_response["motif"] is not None:
                        self.motif += temp_response["motif"]
        return


    def get_direct_cause(self,client,variant):
        ## first figure out if it's in the bad mutation type
        self.get_mutation_type(client,variant)
        ## if it is, get the gene
        ## if it's not, get the tad genes
        if not self.var_is_bad:
            self.get_regulatory_elements_for_var()
        return()


    def get_eqtl_for_genes(self):
        client = EnsemblRestClient()
        for gene in self.genes:
            endpoint = "/eqtl/variant_name/homo_sapiens/"+self.id+"?stable_id="+gene+";statistic=p-value"
            temp_locus_response = client.perform_rest_action(endpoint)#, params = "feature=gene")
            self.eqtls[gene] = temp_locus_response
        return()

    #############################################3
    ## set the strings

    def get_reg_str(self,r):
        return(r["description"]+","+str(r["id"])+","+str(r["seq_region_name"])+":"+str(r["start"])+"-"+str(r["end"]))


    def get_other_reg_str(self,r):
        return(r["description"]+","+str(r["so_accession"])+","+str(r["seq_region_name"])+":"+str(r["start"])+"-"+str(r["end"]))

    def get_motif_str(self,r):
        return(r["transcription_factor_complex"]+","+str(r["stable_id"])+","+str(r["seq_region_name"])+":"+str(r["start"])+"-"+str(r["end"]))

    def get_str_for_reg_element(self,r):
        if r["feature_type"] == "regulatory":
            return(self.get_reg_str(r))
        if r["feature_type"] == "other_regulatory":
            return(self.get_other_reg_str(r))
        if r["feature_type"] == "motif":
            return(self.get_motif_str(r))
        else:
            for key in sorted(r.keys()):
                print(key)
            print(r)
            sys.exit()

    def get_regulatory_str(self):
        ## set the defaults
        reg_str = "NA"
        other_reg_str = "NA"
        motif_str = "NA"

        ## get the regulatory str
        if self.regulatory != []:
            temp_reg_list = []
            for reg_dict in self.regulatory:
                temp_reg_list.append(self.get_str_for_reg_element(reg_dict))
            ## sometimes tehre are duplicates
            temp_reg_list = sorted(list(set(temp_reg_list)))
            reg_str = '|'.join(temp_reg_list)

        ## get the other_regulatory
        if self.other_regulatory != []:
            temp_other_reg_list = []
            for other_reg_dict in self.other_regulatory:
                temp_other_reg_list.append(self.get_str_for_reg_element(other_reg_dict))
            temp_other_reg_list = sorted(list(set(temp_other_reg_list)))
            other_reg_str = '|'.join(temp_other_reg_list)

        ## get the motif strings
        if self.motif != []:
            temp_motif_list = []
            for reg_dict in self.motif:
                temp_motif_list.append(self.get_str_for_reg_element(reg_dict))
            temp_motif_list = sorted(list(set(temp_motif_list)))
            motif_str = '|'.join(temp_motif_list)        
        
        ## Join the results and return them
        return('\t'.join([reg_str, other_reg_str, motif_str]))

    def get_locus_str(self):
        locus_out_list = [] 
        for temp_locus in self.mapped_loci:
            if temp_locus["contig"][:3]=="chr":
                pass
            else:
                locus_out_list.append(temp_locus["contig"]+":"+str(temp_locus["start"])+"-"+str(temp_locus["end"]))
        locus_out_list = sorted(locus_out_list)
        return("|".join(locus_out_list))

    def get_tad_str(self):
        tad_list = []
        for tad in self.tad_annotations:
            tad_list.append(tad[0]+":"+str(int(tad[1]))+"-"+str(int(tad[2]))+","+tad[3])
        return("|".join(tad_list))

    def get_ensembl_link(self):
        return("https://useast.ensembl.org/Homo_sapiens/Location/View?db=core;source=dbSNP;v="+self.id+";vdb=variation")

    def __str__(self):
        temp_str_list = [self.id,self.get_locus_str(),','.join(deepcopy(self.nearest_genes)),','.join(deepcopy(self.genes)), self.get_tad_str(), self.mutation_type, self.get_ensembl_link(), self.get_regulatory_str()]
        temp_str_list = list(map(str,temp_str_list))
        return("\t".join(temp_str_list))
####################################################################################


def get_genes_from_locus(contig, start, end):
    client = EnsemblRestClient()
    genes = []
    if contig[:3] == "chr":
        contig = contig[3:]
    endpoint = "/overlap/region/human/"+contig+":"+str(int(float(start)))+".."+str(int(float(end)))+"?feature=gene"
    temp_locus_response = client.perform_rest_action(endpoint)#, params = "feature=gene")
    if temp_locus_response == None:
        return([])
    for entry in temp_locus_response:
        if "gene_id" in entry:
            genes.append(entry["gene_id"])
    genes = sorted(list(set(genes)))
    return(genes)


######################################################################################

def get_tad_variants(all_variants, tad_file):
    tad_dict = import_dict(tad_file)
    for variant in all_variants:
        if variant.genes == []:
            ## now we get the genes for the intergenic and intronic
            #print("finding the TAD mapping for",variant)
            for mapping in variant.mapped_loci:
                temp_contig = mapping["contig"]
                if temp_contig in tad_dict:
                    temp_genes, temp_tad_annotations, tad_list =  tad_dict[temp_contig].get_genes_from_bp_site(mapping["start"], mapping["end"])
                    variant.genes += temp_genes
                    if len(temp_tad_annotations)>0:
                        if temp_tad_annotations[0] != None:
                            variant.tad_annotations += temp_tad_annotations
                    if len(tad_list)>0:
                        variant.tads+=tad_list
        print(variant)
        # for mapping in variant.mapped_loci:
        #     print("\t",mapping)
        # print("\tnum tad elements:",len(variant.tad_annotations),variant.tad_annotations)
    return(all_variants)

def summarize_all_variants(all_variants, out_dir):
    header = ["variant","mapped_loci",'nearest_gene',"gene_mappings","TAD_mapping","most_pathologic_impact", "ensembl_viewer_link", "regulatory_annotation","other_regulatory_annotation","motif_annotation"]
    out_text = '\t'.join(header)+"\n"+'\n'.join(list(map(str,all_variants)))
    make_file(out_text,os.path.join(out_dir,"variant_summary.tsv"))
    return


def set_nearest_gene(all_variants,tad_file):
    tad_dict = import_dict(tad_file)
    for variant in all_variants:
        print(variant)
        ## now we get the genes for the intergenic and intronic
        #print("finding the TAD mapping for",variant)
        if variant.var_is_bad:
            print("already have annotations for\n\t",variant)
            pass
            ## if it's a bad mutation, give it the annotation 
            #variant.nearest_genes = deepcopy(variant.genes)
        else:
            for mapping in variant.mapped_loci:
                temp_contig = mapping["contig"]
                if temp_contig in tad_dict:
                    print(temp_contig, mapping["start"], mapping["end"])
                    nearest_gene = tad_dict[temp_contig].get_nearest_gene(mapping["start"], mapping["end"])
                    variant.nearest_genes += nearest_gene
                    variant.nearest_genes = list(set(variant.nearest_genes))
    return(all_variants)



def map_variants_to_genes(var_file, stringdb_dir, out_dir):
    ## check that the tad file is present
    tad_file = os.path.join(stringdb_dir,"tad_dict.pkl")
    if not os.path.isfile(tad_file):
        print("we couldn't find the tad file!\n",tad_file)
        sys.exit()
    ## do the analysis
    print('\nreading variants')
    variants = read_file(var_file,'lines')
    ## start the ensembl restful client
    client = EnsemblRestClient()
    all_variants = []
    for variant in variants:
        print("\n\n",variant)
        all_variants.append(variant_obj(client,variant))
    all_variants = get_tad_variants(all_variants = all_variants, 
                                    tad_file = tad_file)
    all_variants = set_nearest_gene(all_variants, tad_file)
    summarize_all_variants(deepcopy(all_variants), out_dir)

    return(all_variants)






#####################################################################

#####################################################################

def main(args):
    map_variants_to_genes(var_file = args.infile, 
                          stringdb_dir = args.stringdb_dir, 
                          out_dir = args.out)
    return

if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    ## global arguments
    parser.add_argument(
        '-infile','-in','-i',
        dest='infile',
        help="the text file with an rsID list",
        type=str)

    parser.add_argument(
        "-stringdb_dir",'-sdb',
        help='The directory containing the StringDB action files',
        type = str,
        default = '/usr/local/lib/cell_signals/')

    parser.add_argument(
        '-out','-o',
        help="the output file containing the rsID mapping ",
        type=str)

    args = parser.parse_args()
    main(args)


#!/usr/bin/env python3

import fileinput, h5py, argparse, os, sys, scipy
import numpy as np
from scipy.sparse import csc_matrix, lil_matrix

parser = argparse.ArgumentParser()
parser.add_argument('infile', type=str, help = "this must be a tab delimited file which will be converted to the hdf5 format (output will be created with the same filename with the .hdf5 extention)")
<<<<<<< HEAD
# parser.add_argument('-sparse', 
# 	                action="store_true",
# 	                help = "add this argument if you want to make it a sparse (scipy csc format)")
=======
parser.add_argument('-sparse', 
	                action="store_true",
	                help = "add this argument if you want to make it a compressed matrix. Takes up less space, but is slower")
>>>>>>> d362fc61e0a62383450b5d042be8c77c99e4bbc6

args = parser.parse_args()

args.infile = os.path.abspath(args.infile)
# if args.sparse:
# 	import h5sparse as h5py
###################################################
## first read through the input tab delim 
## file to see the dimentions of the input matrix
print('reading in the file, please wait')
line_count = 0
cols = None
col_names = None
ID_list = [] ## the variable names (the length here will also be the number of rows in the .h5 matrix)
for line in fileinput.input(args.infile):
	line = line.strip()
	line = line.split('\t')
	if line_count % 1000==0:
		print('\tline',line_count)
	if line_count==0:
		## note that the number of columns for the
		## hdf5 file will be cols-1 because the first
		## column in the input file will be the 
	    cols = len(line)
	    col_names = line
	else:
		if cols!=len(line):
			sys.exit('there is a line ('+str(line_count)+') has an inconsistent number of rows with previous rows')
		ID_list.append(line[0])
		
		
	line_count += 1

fileinput.close()

##########################

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

temp=str(args.infile).split('/')
temp=('/').join(temp[:-1])

make_file('\n'.join(col_names),temp+'/column_IDs.txt')
make_file('\n'.join(ID_list),temp+'/ID_list.txt')


##########################

print('we detected '+str(len(ID_list))+'\t'+'data rows')
print('   and      '+str(cols-1)+'\t'+'data cols')

print()
###################################################

print("now we will make the hdf5 file and populate it with the values from your input dataset")

outfile = os.path.splitext(args.infile)[0]+'.hdf5'
f = h5py.File(outfile, "w")

## set up the data matrix (this assumes float32)
if False:#args.sparse:
	dset = lil_matrix((len(ID_list),cols-1))
	#dset = csc_matrix((len(ID_list),cols-1))
	#dset = f.create_dataset("infile", (len(ID_list),cols-1), dtype=csc_matrix)
else:
<<<<<<< HEAD
	dset = f.create_dataset("infile",
		                    (len(ID_list),cols-1),
		                    dtype=np.float32,
		 			        fillvalue=0.,
                            compression='gzip',
                            compression_opts=9)
=======
	if args.sparse:
		dset = f.create_dataset("infile",
			                    (len(ID_list),cols-1),
			                    dtype=np.float32,
			 			        fillvalue=0.,
	                            compression='gzip',
	                            compression_opts=9)
	else:
		dset = f.create_dataset("infile",
			                    (len(ID_list),cols-1),
			                    dtype=np.float32)
>>>>>>> d362fc61e0a62383450b5d042be8c77c99e4bbc6

## populate dset with the values from the input dataset
line_count = 0
for line in fileinput.input(args.infile):
	line = line.strip()
	line = line.split('\t')
	if line_count % 1000==0:
		print('\tline',line_count)
	if line_count==0:
	## the first line should be column headers, so it is not included
		pass
	else:
		dset[(line_count-1),:] = np.array(list(map(np.float32, line[1:])))
		
		
	line_count += 1

## close the input file
fileinput.close()

## close the hdf5 file
if False:#args.sparse:
	f["infile"]=csc_matrix(dset)
f.close()


################


## validate the hdf5 file
#h5f = h5py.File(outfile, 'r')
#
#print(h5f["infile"][0])













#!/usr/bin/env python3
##import dependency libraries
import sys,time,glob,os,pickle,fileinput,argparse
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
#import pandas as pd
##############################################################
## basic function library
def read_file(tempFile,linesOraw='lines',quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def get_file_path(in_path):
    in_path = in_path.split('/')
    in_path = in_path[:-1]
    in_path = '/'.join(in_path)
    return(in_path+'/')


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)
    

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)

#######################################################

parser = argparse.ArgumentParser()



parser.add_argument(
    '-base_dir','-in','-i','-input',
    dest='base_dir',
    type=str)


args = parser.parse_args()
########################################################
########################################################
if args.base_dir[:-1]!='/':
    args.base_dir+='/'

out_file = args.base_dir + "PyMINEr_summary.html"
base_dir = args.base_dir

if base_dir[-1]!='/':
    base_dir+='/'


########################################################
############# add the various elements #################
from PIL import Image
def get_img_dims(img):
    image = Image.open(img)
    w, h = image.size
    ## reduce size so it fits on the screne
    reduction_factor = w/600
    return(int(w/reduction_factor), int(h/reduction_factor))

def add_h2(title):
    return("\n\t<h2>"+str(title)+"</h2>\n")

def add_h3(title):
    return("\n\t<h3>"+str(title)+"</h3>\n")

def add_p(p):
    return('\t\t\t<p>'+str(p)+'</p>\n')

def add_file_link(f,name):
    return('<a href='+str(f)+'>'+str(name)+'</a>\n')

def add_file_link_list(f,name):
    return('<li><a href='+str(f)+'>'+str(name)+'</a></li>\n')

def add_br():
    return('<br></br>\n')

def add_img(img,alt, base_dir= None):
    if os.path.isfile(img):
        w, h = get_img_dims(img)
        return('<div><img src="'+str(img)+'" alt="'+str(alt)+'" width="'+str(w)+'" height="'+str(h)+'"></div>\n')
    else:
        if base_dir != None:
            curr_dir = os.getcwd()
            os.chdir(base_dir)
            final_text = add_img(img, alt)
            os.chdir(curr_dir)
            return(final_text)
        print("\tWARNING: couldn't find:"+img)
        return(add_br())

def add_button_head(text):
    return("""<button class="collapsible">"""+str(text)+"""</button>
<div class="content">""")



def add_tab_header(section_name, default_open=False):
    out_head = "<div id='"+section_name+"' class='w3-container w3-border city'"
    if not default_open:
        out_head += ' style="display:none"'
    else:
        out_head += ' style="display: block;"'
    out_head += '>'
    return(out_head)



def add_table(in_table):
    ## start the table
    out_table_str = """<table style="width:100%">\n"""
    for i in range(len(in_table)):
        out_table_str+="\t<tr>\n"
        for j in range(len(in_table[i])):
            out_table_str+="\t<th>"+str(in_table[i][j])+"</th>\n"
        out_table_str+="\t</tr>\n"
    out_table_str+="</table>\n"
    return(out_table_str)


def parse_file_name(img):
    old_name = img[:-4]
    temp_name = old_name.split('_')
    temp_name = ' '.join(temp_name)
    return(old_name,temp_name)



def add_clustering(base_dir):
    os.chdir(base_dir)
    cluster_str = ""
    cluster_str+=add_tab_header("Clustering")
    #cluster_str+=add_h2('Clustering')
    #cluster_str+=add_button_head('Clustering')
    ## check if anti_correlation clustering was done
    if os.path.isfile(base_dir+'sig_neg_count_vs_total_neg_count.png'):
        ## if it is, start annotating
        cluster_str+=add_h3("Negative Control Bootstrap Shuffling")
        cluster_str+=add_p("Here, PyMINEr shuffled up your data to make it randomized. This will maintain the overall distribution of your data, while at the same time randomizing it so that we can come up with a reasonable cutoff for performing the anti-correlation based feature selection. Shown below is the distribution of all randomized Spearman rhos.")
        cluster_str+=add_img("boostrap_cor_rhos.png", 'boot_all')
        cluster_str+=add_p("Here are just the negative correlations from the shuffled up dataset.")
        cluster_str+=add_img("boostrap_neg_cor_rhos.png", 'boot_neg')
        cluster_str+=add_p("Here is a scatter plot that shows the number of total negative correlations observed for each gene (y-axis), and the log2 number of significant negative correlations (x-axis). All of the genes to the left/below the black line were used for clustering.")
        cluster_str+=add_img("sig_neg_count_vs_total_neg_count.png","sig_vs_total_neg")
        cluster_str+=add_p("Similarly, here is a plot showing the ratio of significant to non-sigificant. Everything to the right was used for clustering.")

    ## link to the file with all of the sample annotations
    cluster_str+=add_h3("Clustering Results:")
    cluster_str+=add_file_link("sample_clustering_and_summary/sample_k_means_groups.txt",'sample group annotations')

    ## get the images from the sample_clustering_and_summary folder
    os.chdir(base_dir+"sample_clustering_and_summary/")
    additional_images=[]
    for item in glob.glob("*.png"):
        additional_images.append(item)
    os.chdir(base_dir)
    if len(additional_images)>0:
        cluster_str+=add_h3('Here are some additional images:')
        for img in additional_images:
            old_name = img[:-4]
            temp_name = old_name.split('_')
            temp_name = ' '.join(temp_name)
            cluster_str+=add_p(temp_name)
            cluster_str+=add_img("sample_clustering_and_summary/"+img,old_name)
    
    static_files = []
    cluster_str+='\n</div>\n'
    cluster_str+='\n</div>\n'
    os.chdir(base_dir)
    return(cluster_str)



def add_statistics(base_dir):
    stats_str = ""
    stats_str+=add_tab_header("Differential Expression")
    #stats_str+=add_h2("Statistics")
    #stats_str+=add_button_head("Statistics")
    stats_str+=add_h3("basic stats")
    stats_str+=add_p("If you find some interesting genes that are different between groups, here are the:")
    stats_str+=add_file_link("sample_clustering_and_summary/k_group_means.txt",'group means')
    stats_str+=add_p('and...')
    stats_str+=add_file_link("sample_clustering_and_summary/k_group_sd.txt",'group standard deviations')
    stats_str+=add_p('as well as the')
    # stats_str+=add_file_link("sample_clustering_and_summary/sample_var_enrichment_Zscores.txt",'sample-wise Z-scores')
    # stats_str+=add_p('and...')
    stats_str+=add_file_link("sample_clustering_and_summary/k_group_enrichment.txt",'group level Z-scores')
    stats_str+=add_h3("statistical comparisons")
    sig_dir = "sample_clustering_and_summary/significance/"

    stats_str+=add_file_link(sig_dir+'groups_1way_anova_results.txt',"Benjamini-Hoschberg corrected 1-way Anovas")
    stats_str+=add_p("For just getting a high-level view of the differences between cell-types, the pathway analysis of these results may be more interesting. Those are a couple tabs down. Looking at the gene expression module usage and the pathways associated with those modules can also be very useful for figuring out the functions of the genes being expressed in a cluster.")
    stats_str+='\n</div>\n'
    stats_str+='\n</div>\n'
    return(stats_str)

def add_gene_enrichment(base_dir):
    os.chdir(base_dir)
    sig_dir = "sample_clustering_and_summary/significance/"
    enrich_str = ""
    enrich_str+=add_tab_header("Cluster Pathway Analysis")
    #enrich_str+=add_h2("Gene Enrichment in Groups")
    #enrich_str+=add_button_head("Gene enrichment in groups")
    enrich_str+=add_h3("Significantly enriched genes in each group")
    enrich_str+=add_p("Below is a file that contains a table with genes on the left, and groups in columns. If a gene is considered significantly enriched, that means that the BH corrected 1-way Anova was significant, and the gene had a high Z-score in that particular group. If a gene is significantly enriched in that group, the value is True in the corresponding cell in the table, False if it was not significantly enriched.")
    enrich_str+=add_file_link(sig_dir+"/significant_and_enriched_boolean_table.txt","True/False significantly enriched table")
    enrich_str+=add_p("alternatively, you can use these separated files that simply have the list of significantly enriched genes for each group:")
    enrich_str+=add_button_head('Genes enriched in each group')
    os.chdir(sig_dir)
    enrich_files=[]
    for item in glob.glob("*_significant_enriched.txt"):
        enrich_files.append(item)
    enrich_files = sorted(enrich_files)
    if len(enrich_files)>0:
        enrich_str+="\t\t\t\t<ul>\n"
        for img in enrich_files:
            old_name = img[:-4]
            temp_name = old_name.split('_')
            temp_name = ' '.join(temp_name)
            #enrich_str+=add_p(temp_name)
            enrich_str+=add_file_link(sig_dir+img,old_name)
            enrich_str+=add_br()

        enrich_str+="\t\t\t\t</ul>\n"

    enrich_str+="\t\t</div>"# end the sub-button
    enrich_str+=add_br()

    os.chdir(base_dir)
    pathway_dir = sig_dir+'gprofiler/'
    os.chdir(pathway_dir)
    enrich_path=[]
    for item in glob.glob("*.txt"):
        enrich_path.append(item)
    os.chdir(base_dir)
    enrich_path = sorted(enrich_path)
    if len(enrich_path)>0:
        #enrich_str+=add_h3('Pathway Enrichment for Each Group:')
        enrich_str+=add_button_head('Pathway enrichment for each group')
        if os.path.isfile(sig_dir+"/combined_neg_log10p_gprofiler.txt"):
            enrich_str+=add_p("Below is a combined file with all of the pathways that came out of the analysis of the above genes, enriched in the different groups. We've also created a new algorithm for ranking the importance of these pathways (not to boast, but I'm kind of proud of it). It's based on the information/entropy of the -log10(p-vals). If you look at the -log10(p-vals) at the top of the list, you should find pathways that are really high in some groups and really low in other groups. The overall formula is calculated by taking the sum(KL-divergance)*range(-log10(p-vals)). The KL-divergence is looking for a difference in the distribution between the observed -log10(p-vals) and the distribution expected from an uninformative vector of -log10(p-vals) (this is a Gaussian null hypothesis). Then we multiply the sum(KL-divergance) by the range of -log10(p-vals) to let the most significant pathways with lots of information/low entropy rise to the top.")
            enrich_str+=add_file_link(sig_dir+"/combined_neg_log10p_gprofiler.txt",'combined pathway analyses')
            enrich_str+=add_p("and here is a file that has a normalized metric that ranks each pathway for their individual importance within each group. It would be useful to sort each of these and see what rises to the top for each 'cell type' or whatever your groups are.")
            enrich_str+=add_file_link(sig_dir+'/individual_class_importance.txt', 'individual class importance')
            enrich_str+=add_p("Below are all of the individual results so that you find which genes were in which pathways in individual groups:")
            enrich_str+=add_button_head("individual pathway files")
        enrich_str+="\t\t\t\t<ul>\n"
        for img in enrich_path:
            old_name = img[:-4]
            temp_name = old_name.split('_')
            temp_name = ' '.join(temp_name)
            #enrich_str+=add_p(temp_name)
            enrich_str+=add_file_link(pathway_dir+img,old_name)
            enrich_str+=add_br()


        if os.path.isfile(sig_dir+"/combined_neg_log10p_gprofiler.txt"):
            enrich_str+="\n\t\t</div>\n"

        enrich_str+="\t\t\t\t</ul>\n"
        enrich_str+="\n\t\t</div>\n"#end the sub-button
        enrich_str+=add_br()

    enrich_str+='\n</div>\n'
    return(enrich_str)


def add_community_pathway_annotation(com, individual_class_importance_table):
    if type(individual_class_importance_table) != np.ndarray:
        return(add_br())
    print(individual_class_importance_table)
    ## get the top 5 pathways
    ## first get the right column
    header = individual_class_importance_table[0,:]
    temp_col = np.where(header == com+"_gprofiler.txt")[0]
    print(temp_col)
    print(com,individual_class_importance_table[1:,temp_col])
    most_sig_order = np.argsort(np.array(individual_class_importance_table[1:,temp_col].tolist(),dtype=np.float32), axis = 0)[::-1] + 1
    #print(individual_class_importance_table[most_sig_order,temp_col])
    #print(most_sig_order)
    keep_rows = []
    for i in range(0,5):
        temp_row = most_sig_order[i]
        ## check if it's actually significant first
        print(individual_class_importance_table[temp_row,temp_col])
        if float(individual_class_importance_table[temp_row,temp_col]) > 0:
            keep_rows.append(temp_row)
    ## go through the rows that we're keeping & summarize them
    out_table = [["term_id","term_def","importance_metric"]]
    for i in keep_rows:
        temp_line = [individual_class_importance_table[i,1][0],individual_class_importance_table[i,3][0],individual_class_importance_table[i,temp_col][0]]
        out_table.append(temp_line)
    return(add_table(out_table)+add_br())


def add_community_top_markers(base_dir, com, community_annotations_file, top_n_markers = 15):
    com_marker_str = ""
    com_marker_str += add_button_head("top marker genes of "+com+" (defined by high Local PageRank)")
    ## read in the community annotations file
    try:
        com_anno_table = read_table(community_annotations_file)
    except:
        cur_dir = os.getcwd()
        os.chdir(base_dir)
        com_anno_table = read_table(community_annotations_file)
        os.chdir(cur_dir)
    else:
        com_anno_table = read_table(community_annotations_file)
    num_lines = min([len(com_anno_table),top_n_markers])
    temp_table = com_anno_table[:num_lines]
    # for i in range(len(temp_table)):
    #     print(temp_table[i])

    com_marker_str += add_p("this is a list of the top genes that are most central to this hub, and will therefore likely make good marker genes.")
    com_marker_str += add_table(temp_table)
    # print(add_table(temp_table))
    # sys.exit()
    com_marker_str += "</div>"## ends the button tab


    return(com_marker_str)



def add_community_str(base_dir, com, individual_class_importance_table):
    com_str = ""
    com_str += add_h3("Community module usage")
    com_str += add_community_pathway_annotation(com, individual_class_importance_table)
    com_dir = os.path.join(os.path.join("pos_cor_graphs/community_analysis/",com))
    community_annotations_file = os.path.join(com_dir,"community_ids_annotated.tsv")
    com_str += add_community_top_markers(base_dir, com, community_annotations_file)
    com_str += add_file_link(community_annotations_file,"This is the list of genes in the community.")
    com_str += add_br()
    com_str += add_file_link(os.path.join(com_dir,"TukeyHSD.tsv"),"This is the differential module usage across groups")
    com_str += add_p("And here is a plot showing the module usage across groups.")
    com_str += add_img(os.path.join(com_dir,"group_z_scores.png"),"Group Z-scores", base_dir = base_dir)
    com_str += add_br()
    com_str += add_p("This is where this community is in the network graph.")
    com_str += add_img(os.path.join(com_dir,"community.png"),"community subset", base_dir = base_dir)
    return(com_str)



def add_graph(base_dir):
    graph_str=""
    #graph_str+=add_h2("Expression graphs")
    #graph_str+=add_button_head("Expression graphs")
    graph_str+=add_tab_header("Gene Expression Networks")
    graph_str+=add_h3("Adjacency Lists:")
    adj_list_list = []
    os.chdir(base_dir)
    for f in glob.glob("*"):
        if 'adj_list' in f:
            adj_list_list.append(f)
    print(adj_list_list)
    for f in adj_list_list:
        if '_pos.tsv' in f:
            pos_adj = f
        elif '_neg.tsv' in f:
            neg_adj = f
        else:
            total_adj = f
    # graph_str+=add_p('Full adjacency list:')[:-6]+'</p>'
    # print(base_dir+total_adj)
    graph_str+="\t\t\t\t<ul>\n"
    graph_str+=add_file_link(total_adj,'Full adjacency list')
    graph_str+=add_br()
    graph_str+=add_file_link(pos_adj,'Positive correlation (co-expression) adjacency list')
    graph_str+=add_br()
    graph_str+=add_file_link(neg_adj,'Negative correlation adjacency list')
    graph_str+=add_br()
    graph_str+="\t\t\t\t</ul>\n"

    graph_str+=add_h2("Co-expression graph plots:")
    coexpression_graph_dir = "pos_cor_graphs/"
    graph_str+=add_h3("This is a 'hairball' view of a Spearman coexpression network of your data, where each point represents a gene, and the lines are whether or not a gene is correlated with another gene. The closer two genes are to each other, the more correlated they are.")
    graph_str+=add_img(coexpression_graph_dir+"full_graph.png",'Co-expression graph')
    if os.path.isfile(coexpression_graph_dir+"community.png"):
        graph_str+=add_h3("This is a graph of all the communities of genes that are coordinately regulated in your dataset")
        graph_str+=add_img(coexpression_graph_dir+"community.png",'communities')

    ## 
    community_dir = os.path.join(coexpression_graph_dir,"community_analysis")
    if os.path.isdir(community_dir):
        graph_str+=add_button_head('Analysis of individual communities')
        os.chdir(community_dir)
        ## first add the significance and pathway files
        graph_str+=add_br()
        graph_str+=add_file_link(os.path.join(base_dir, community_dir, "global_statistics.tsv"),"Here are the stats (BH corrected 1-way anova) for differential module usage among groups.")
        graph_str+=add_br()
        graph_str+=add_file_link(os.path.join(base_dir, community_dir, "combined_neg_log10p_gprofiler.txt"),"-log10(p-val) for the pathways associated with each module")
        graph_str+=add_br()
        individual_class_importance_file = os.path.join(base_dir, community_dir, "individual_class_importance.txt")
        graph_str+=add_file_link(individual_class_importance_file,"The unique pathways of each module (individual class importance).")
        graph_str+=add_br()
        ## second add the folders for each individual community
        print(os.getcwd())
        community_dirs = []
        for file in glob.glob('*'):
            if os.path.isdir(file):
                if file != 'gprofiler':
                    print(file)
                    community_dirs.append(file)

        ## catelogue how many genes are in each community
        com_num_dict = {}
        com_num_list = []
        for com in community_dirs:
            com_num_dict[com] = len(read_file(os.path.join(com,"community_ids.txt"),"lines"))
            com_num_list.append(com_num_dict[com])

        ## sort based on size
        new_order = np.argsort(np.array(com_num_list))[::-1]
        community_dirs = np.array(community_dirs)[new_order].tolist()
        if os.path.isfile(individual_class_importance_file):
            individual_class_importance_table = np.array(read_table(individual_class_importance_file))
        else:
            individual_class_importance_table = "None"
        for com in community_dirs:
            graph_str+=add_button_head(com+" ("+str(com_num_dict[com])+" nodes)")
            graph_str+=add_community_str(base_dir, com, individual_class_importance_table)
            graph_str+="\n\t\t</div>\n"
            #graph_str+=add_br()


        #graph_str+="\n\t\t</div>\n"
        graph_str+=add_br()
        graph_str+="\n\t</div>\n"
        os.chdir('../..')
        print(os.getcwd())

    if os.path.isfile(coexpression_graph_dir+"PageRank.png"):
        graph_str+=add_h3("PageRank is a metric for how well connected a gene is in the coexpression network")
        graph_str+=add_img(coexpression_graph_dir+"PageRank.png",'Page-Rank')
    if os.path.isfile(coexpression_graph_dir+"LPR.png"):
        graph_str+=add_h3("Local PageRank is a derivative of both Louvain-modularity based community detection and PageRank. It essentially calculates a normalized PageRank within each module to quantify local connectivy throughout the graph.")
        graph_str+=add_img(coexpression_graph_dir+"LPR.png",'Local PageRank')
    graph_str+=add_h3("Here are the Z-scores for each group overlaid on the co-expression graph (red means it's enriched in that group, blue means it's low expression or not expressed):")

    ## find all of the pertinent files, and
    graph_str+=add_button_head('Z-score overlaid co-expression graphs')
    os.chdir(coexpression_graph_dir)
    extra_images = []
    for file in glob.glob('*.png'):
        if 'sample_group' in file:
            extra_images.append(file)
    extra_images = sorted(extra_images)
    os.chdir(base_dir)
    for img in extra_images:
        new,old = parse_file_name(img)
        graph_str+=add_p(new)
        graph_str+=add_img(coexpression_graph_dir+img, new)
    graph_str+="\n\t\t</div>\n"
    #graph_str+=add_br()
    graph_str+="\n\t</div>\n"
    graph_str+="\n\t</div>\n"
    return(graph_str)


def add_img_list(list_of_images,temp_dir):
    output = ""
    for img in list_of_images:
        new,old = parse_file_name(img)
        output+=add_p(new)
        output+=add_img(temp_dir+img, new)
    return(output)


def add_genes_of_interest(base_dir):
    goi_str=""
    goi_dir = 'genes_of_interest/'
    goi_str+=add_tab_header('Genes Of Interest')
    if os.path.isdir(goi_dir):
        #goi_str+=add_button_head('Genes of interest')
        goi_str+=add_p("An interesting way to use the structure of the graphs generated by PyMINEr is looking at how far away all genes in your dataset are away from your genes of interest. We've shown in the PyMINEr paper that there are many types of functional enrichment that correlate with how far away a gene is from another gene in the graph network. For example, close to a transcription factor are more likely to have a binding site for that transcription factor when compared to a gene that's father away from it. There is also an increased probability that two genes will encoded proteins that have a physical intereaction when those two genes are directly connected (i.e. 1-degree of separation).")
        goi_str+=add_p('below is a file containing a table that has the distance of all genes in the genome away from your gene(s) of interest.')
        goi_str+=add_file_link(goi_dir+'genes_of_interest_shortest_path_list.txt','Shortest path of all genes away from your genes of interest')
        goi_str+=add_p('Below is a heatmap of your genes of interest:')
        goi_str+=add_img(goi_dir+'genes_of_interest_subset_heatmap.png','genes of interest heatmap')

        ## collect the other images
        os.chdir(goi_dir)
        additional_images = []
        for f in glob.glob('*.png'):
            if f != 'genes_of_interest_subset_heatmap.png':
                additional_images.append(f)
        goi_str+=add_button_head("additional plots for your genes of interest")
        os.chdir(base_dir)
        goi_str+=add_img_list(sorted(additional_images),goi_dir)
        goi_str+='\n\t\t</div>\n'

        goi_str+="\n</div>\n"
    else:
        goi_str+=add_p("We didn't find any genes of interest to analyze. For future usage - if you do want to provide a list of genes to make some special plots for, you can include a text file that has the names of the genes, with the option of a second column for an alias. Provide that file after the -goi argument.")
    goi_str+="\n</div>\n"
    return(goi_str)


def add_autocrine_paracrine(base_dir):
    ap_dir = "autocrine_paracrine_signaling/"
    ap_str=""
    ap_str += add_tab_header("Autocrine/Paracrine Signaling")
    if not os.path.isdir(ap_dir):
        add_p("Something went wrong & we didn't get any autocrine and paracrine signaling results. There might have been an error within the run.")
    else:
        #ap_str+=add_button_head("Autocrine/paracrine signaling")
        ap_str+=add_p("Below are the predicted signaling networks")
        ap_str+=add_file_link(ap_dir+'all_cell_cell_interaction_summary.txt','Number of interactions between all of the groups')
        ap_str+=add_br()
        ap_str+=add_file_link(ap_dir+'all_cell_type_specific_interactions.txt','A detailed summary of each autocrine/paracrine interaction')
        ap_str+=add_br()
        ap_str+=add_file_link(ap_dir+'all_cell_type_specific_interactions_gprofiler.txt','A detailed summary of the pathways signaling across and within groups')
        ap_str+=add_br()
        ap_str+=add_file_link(ap_dir+'combined_neg_log10p_gprofiler.txt','A table of negative log10 p-values that for each pathway in each interaction. (Zero just means it did not reach signficance)')
        ap_str+=add_br()
        ap_str+=add_file_link(ap_dir+'individual_class_importance.txt','A table of the individual importance of each pathway for the given group')
    ap_str += "\n</div>\n"
    return(ap_str)
        #combined_neg_log10p_gprofiler.txt
        #individual_class_importance.txt


def add_gene_annotations(base_dir):
    anno_str=""
    anno_str+=add_tab_header("Gene Annotations",default_open=True)
    anno_str+=add_h3("Genes Annotations")
    if os.path.isfile("annotations.tsv"):
        anno_str+=add_p("Here are the annotations for your genes. You can use this in excel using v-lookup if you want to get gene symbols or definitions for any of the other files.")
        anno_str+=add_file_link("annotations.tsv","Annotation file")
    if os.path.isfile("human_orthologues.tsv"):
        anno_str+=add_br()
        anno_str+=add_file_link("human_orthologues.tsv","Human orthologues to your genes")
    anno_str+=add_h3("Clustering")
    anno_str+=add_p("The first step as identifying cell-type/-state clusters. This process includes feature selection followed\
     by clustering (usually either by locally weighted Louvain-modularity, or affinity propagation depending on the arguments that you used). The clustering tab has some plots that show traditional things like PCA/UMAP, etc")
    anno_str+=add_h3("Differential Expression")
    anno_str+=add_p("This section holds links to .tsv format tables that run the statistics \
        that compare the clusters \
        to each other. Plots are on other tabs, this one is just tables.")
    anno_str+=add_h3("Genes Of Interest")
    anno_str+=add_p("If you gave a file to the -goi argument that holds some genes of interest, \
        the plots generated for those genes of interest are included in this section.")
    anno_str+=add_h3("Cluster Pathway Analysis")
    clust_str="""Tables that show the pathway analysis results, based on """
    clust_str+="""<a href="https://www.ncbi.nlm.nih.gov/pubmed/17478515" target="_blank">gProfiler</a>"""
    clust_str+=" as well as PyMINEr's custom information theory based prioritization of those pathways."
    anno_str+=add_p(clust_str)
    anno_str+=add_h3("Unique Marker Genes")
    anno_str+=add_p("")
    anno_str+=add_h3("Gene Expression Networks")
    anno_str+=add_p("")
    anno_str+=add_h3("Autocrine/Paracrine Signaling")
    anno_str+=add_p("")
    anno_str+='\n</div>\n'
    return(anno_str)

def add_high_marker_genes(base_dir):
    marker_str = ""
    marker_dir = "sample_clustering_and_summary/significance/high_markers/"
    marker_str = ""
    marker_str += add_tab_header("Unique Marker Genes")
    if not os.path.isfile(marker_dir+'marker_gene_annotations.tsv'):
        marker_str+=add_p("No marker genes found")
    else:
        #marker_str+=add_button_head("Highly expressed group-specific markers")
        marker_str+=add_p("PyMINEr analyzes the mean expression of each sample group and then looks for genes that meet three criteria.")
        #marker_str+=add_br()
        marker_str+=add_p("\t1) The gene is significant by 1-way ANOVA (after BH correction, alpha=0.05).")
        #marker_str+=add_br()
        marker_str+=add_p("\t2) The distance between the group with highest mean expression and second highest expression is calculated. If a gene is in the top 90th percentile of this calculation, it can make it through.")
        #marker_str+=add_br()
        marker_str+=add_p("\t3) A metric called the q-value is calculated (usually to identify outliers). The q-value is the ratio of the value calculated in number 2 compared to the range of sample group means. It's essentially what percent of the range is attributable to the distance between the highest expressing group vs the second highest expressing group.")
        #marker_str+=add_br()
        marker_str+=add_p("If all three of these criteria are, met you'll find it with an annotated group in the file below. Note that if you have several very closely related groups, there might not be many highly expressed genes that are exclusivly expressed in an individual group. In this case, you might need several markers at once to descriminate them.")
        #marker_str+=add_br()
        marker_str+=add_file_link(marker_dir+'marker_gene_annotations.tsv',"Highly expressed marker genes")
        marker_str+=add_img(marker_dir+'genes_of_interest_subset_heatmap.png',"High expression marker genes")
        marker_str+='\n</div>\n'
    marker_str+='\n</div>\n'
    return(marker_str)

def add_to_container(container, new_tab_name):
    """"""
    if container[1]=="":
        container[1]="""\n\t<button class="w3-bar-item w3-button tablink w3-red" onclick="openCity(event,'"""+new_tab_name+"""')">"""+new_tab_name+"""</button>\n"""
    else:
        container[1]+="""\n\t<button class="w3-bar-item w3-button tablink" onclick="openCity(event,'"""+new_tab_name+"""')">"""+new_tab_name+"""</button>\n"""
    return(container)


def compile_web_page(header, container, body, tail):

    container = '\n'.join(container)
    final_page = '\n'.join([header, container, body, tail])
    return(final_page)

###########################################################
def pyminer_make_website(base_dir):
    os.chdir(base_dir)

    ## first add the generic heading
    web_header = """<!DOCTYPE html>
    <html>

    <title>PyMINEr Results</title>

    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="https://www.w3schools.com/w3css/4/w3.css">

    <head>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <style>
    .collapsible {
        background-color: #777;
        color: white;
        cursor: pointer;
        padding: 18px;
        width: 100%;
        border: none;
        text-align: left;
        outline: none;
        font-size: 15px;
    }

    .active, .collapsible:hover {
        background-color: #555;
    }

    .content {
        padding: 0 18px;
        display: none;
        overflow: hidden;
        background-color: #f1f1f1;
    }
    table {
      font-family: arial, sans-serif;
      border-collapse: collapse;
      width: 100%;
    }

    td, th {
      border: 1px solid #dddddd;
      text-align: left;
      padding: 8px;
    }

    tr:nth-child(even) {
      background-color: #dddddd;
    }
    </style>
    </head>

    <body>


    <div class="w3-container w3-teal">
       <h1><a href="https://www.sciencedirect.com/science/article/pii/S2211124719300920" target="_blank">PyMINEr</a> Results</h1>
    </div>



    """

    container_text = ["""    <div class="w3-container">
      
      <div class="w3-bar w3-black">""","","""  </div>\n        </div>"""]

    body = ""
    body+=add_gene_annotations(base_dir)
    container_text = add_to_container(container_text, "Gene Annotations")
    body+=add_clustering(base_dir)
    container_text = add_to_container(container_text, "Clustering")
    body+=add_statistics(base_dir)
    container_text = add_to_container(container_text, "Differential Expression")
    body+=add_genes_of_interest(base_dir)
    container_text = add_to_container(container_text, "Genes Of Interest")
    body+=add_gene_enrichment(base_dir)
    container_text = add_to_container(container_text, "Cluster Pathway Analysis")
    body+=add_high_marker_genes(base_dir)
    container_text = add_to_container(container_text, "Unique Marker Genes")
    body+=add_graph(base_dir)
    container_text = add_to_container(container_text, "Gene Expression Networks")
    body+=add_autocrine_paracrine(base_dir)
    container_text = add_to_container(container_text, "Autocrine/Paracrine Signaling")


    tail="""

    <script>
    function openCity(evt, cityName) {
      var i, x, tablinks;
      x = document.getElementsByClassName("city");
      for (i = 0; i < x.length; i++) {
        x[i].style.display = "none";
      }
      tablinks = document.getElementsByClassName("tablink");
      for (i = 0; i < x.length; i++) {
        tablinks[i].className = tablinks[i].className.replace(" w3-red", "");
      }
      document.getElementById(cityName).style.display = "block";
      evt.currentTarget.className += " w3-red";
    }
    </script>



    <script>
    var coll = document.getElementsByClassName("collapsible");
    var i;

    for (i = 0; i < coll.length; i++) {
        coll[i].addEventListener("click", function() {
            this.classList.toggle("active");
            var content = this.nextElementSibling;
            if (content.style.display === "block") {
                content.style.display = "none";
            } else {
                content.style.display = "block";
            }
        });
    }
    </script>



    </body>
    </html>



    """

    out_web = compile_web_page(web_header, container_text, body, tail)


    ###########################################################
    out_web = out_web.replace('//','/')
    out_web = out_web.replace(base_dir,'')

    make_file(out_web,out_file)
    return()



if __name__ == "__main__":
    ####################
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-base_dir','-in','-i','-input',
        dest='base_dir',
        type=str)
    args = parser.parse_args()
    ####################
    base_dir = os.path.join(args.base_dir)


    pyminer_make_website(base_dir)
#cmd('firefox '+out_file)
